@@ -982,8 +982,12 @@ TMP_TABLE_SHARE *THD::create_temporary_table(LEX_CUSTRING *frm,
   saved_key_cache= strmov(tmp_path, path) + 1;
   memcpy(saved_key_cache, key_cache, key_length);
 
+  /*
+    Temp tables can't be thread specific for slaves as they are freed
+    during cleanup() from Relay_log_info::close_temporary_tables()
+  */
   init_tmp_table_share(this, share, saved_key_cache, key_length,
-                       strend(saved_key_cache) + 1, tmp_path);
+                       strend(saved_key_cache) + 1, tmp_path, !slave_thread);
 
   /*
     Prefer using frm image over file. The image might not be available in
@@ -1486,6 +1490,13 @@ bool THD::free_tmp_table_share(TMP_TABLE_SHARE *share, bool delete_table)
   if (delete_table)
   {
     error= rm_temporary_table(share->db_type(), share->path.str);
+
+    if (share->hlindexes())
+    {
+      /* as of now: only one vector index can be here */
+      DBUG_ASSERT(share->hlindexes() == 1);
+      rm_temporary_table(share->hlindex->db_type(), share->hlindex->path.str);
+    }
   }
   free_table_share(share);
   my_free(share);
